# import required packges
import warnings
from collections import defaultdict
import io

warnings.filterwarnings("ignore")


import matplotlib.patches as patches
import numpy as np
from lmfit import Model
from matplotlib import rcParams

# Matplotlib packages
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
)
from matplotlib.backends.backend_qt5agg import (
    NavigationToolbar2QT as NavigationToolbar,
)
from matplotlib.figure import Figure
from matplotlib.path import Path
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtGui import QImage
from PyQt5.QtWidgets import QApplication

from ._settings import settings
from ._tools import float_to_str


rcParams["mathtext.fontset"] = "cm"


class DraggableVLine:
    """class to create a draggable vertical line in a plot"""

    lock = None  # we need this to be able to dragg only one line at a time

    def __init__(self, ax, x, linewidth=4, linestyle="--", color="gray"):
        self.line = ax.axvline(
            x=x, linewidth=linewidth, linestyle=linestyle, color=color
        )
        self.press = None
        self.connect()

    def get_pos(self):
        return self.line.get_xdata()[0]

    def remove(self):
        self.line.remove()

    def connect(self):
        "connect to all the events we need"
        self.cidpress = self.line.figure.canvas.mpl_connect(
            "button_press_event", self.on_press
        )
        self.cidrelease = self.line.figure.canvas.mpl_connect(
            "button_release_event", self.on_release
        )
        self.cidmotion = self.line.figure.canvas.mpl_connect(
            "motion_notify_event", self.on_motion
        )

    def on_press(self, event):
        if event.inaxes != self.line.axes:
            return
        if DraggableVLine.lock is not None:
            return
        contains, _ = self.line.contains(event)
        if not contains:
            return
        x, _ = self.line.get_xdata()
        self.press = x, event.xdata
        DraggableVLine.lock = self

    def on_motion(self, event):
        if self.press is None:
            return
        if DraggableVLine.lock is not self:
            return
        if event.inaxes != self.line.axes:
            return
        x, xpress = self.press
        dx = event.xdata - xpress
        x_clip = x + dx
        self.line.set_xdata([x_clip, x_clip])
        self.line.figure.canvas.draw()

    def on_release(self, event):
        if DraggableVLine.lock is not self:
            return
        DraggableVLine.lock = None
        self.press = None
        self.line.figure.canvas.draw()

    def disconnect(self):
        self.line.figure.canvas.mpl_disconnect(self.cidpress)
        self.line.figure.canvas.mpl_disconnect(self.cidrelease)
        self.line.figure.canvas.mpl_disconnect(self.cidmotion)


class RangeSelector:
    """class that creates a rangeselector in a plot consisting of two draggable vertical lines"""

    def __init__(self, ax, pos1, pos2):
        self.ax = ax  # axes that holds the lines
        self.pos = [pos1, pos2]  # initial positions of the lines
        self.drag_lines = [DraggableVLine(self.ax, pos) for pos in self.pos]

    def get_range(self):
        pos = [dragline.get_pos() for dragline in self.drag_lines]
        pos.sort()
        return pos

    def remove(self):
        for dragline in self.drag_lines:
            dragline.remove()


class PlotWidget(QtWidgets.QWidget):
    """Qt widget to hold the matplotlib canvas and the tools for interacting with the plots"""

    resized = QtCore.pyqtSignal()  # emits when the widget is resized

    def __init__(self, fitter, xlabel, ylabel, **kwargs):
        QtWidgets.QWidget.__init__(self)

        self.setLayout(QtWidgets.QVBoxLayout())
        self.canvas = PlotCanvas(fitter, xlabel, ylabel, **kwargs)
        self.toolbar = NavigationToolbar(self.canvas, self)
        self.toolbar.addSeparator()

        """
        Change Data Button
        """
        self.updateData = QtWidgets.QAction("Update Data")
        self.updateData.setIconText("UPDATE DATA")
        self.updateData.setFont(QtGui.QFont("Times", 12, QtGui.QFont.Bold))
        self.updateData.triggered.connect(self.update_data)

        self.toolbar.addSeparator()

        self.ACshowselector = QtWidgets.QAction("Activate/Clear RangeSelector")
        self.ACshowselector.setIconText("RANGE SELECTOR")
        self.ACshowselector.setFont(QtGui.QFont("Times", 12, QtGui.QFont.Bold))
        self.ACshowselector.triggered.connect(self._toggle_showselector)

        self.toolbar.addSeparator()

        self.toolbar.addAction(self.updateData)
        self.toolbar.addAction(self.ACshowselector)

        self.toolbar.addSeparator()
        self.layout().addWidget(self.toolbar)
        self.layout().addWidget(self.canvas)

        self.resized.connect(
            self.update_plot
        )  # update plot when window is resized to fit plot in window

    def resizeEvent(self, event):
        self.resized.emit()
        return super(PlotWidget, self).resizeEvent(event)

    def update_plot(self):
        self.canvas.update_plot()

    def _toggle_showselector(self):
        self.canvas.toggle_rangeselector()

    def update_data(self):
        return


class PlotCanvas(FigureCanvas):
    """class to hold a canvas with a matplotlib figure and two subplots for plotting data and residuals"""

    def __init__(self, fitter, xlabel, ylabel, **kwargs):
        self.fitter = fitter
        self.data = fitter.data  # contains the x, y and error data
        self.fitline = None  # contains the fitline if available
        self.residuals = None  # contains the residuals if available
        self.complex_residuals = None
        self.initial_guess_line = None
        self.kwargs = kwargs
        self.fitline_kwargs = defaultdict(str)
        self.complex = (
            False if "complex" not in self.kwargs else self.kwargs["complex"]
        )

        # Get Axis titles
        self.ax1_title = (
            self.kwargs["title"] if "title" in self.kwargs else "Data"
        )
        # delete from kwargs to remove matplotlib error
        if "title" in self.kwargs:
            del self.kwargs["title"]
        if "method" in self.kwargs:
            del self.kwargs["method"]
        if "complex" in self.kwargs:
            del self.kwargs["complex"]

        # Separate Fitline kwargs from data kwargs, at this point kwargs and self.kwargs are same thing
        if "fitline_color" in self.kwargs:
            self.fitline_kwargs["color"] = kwargs["fitline_color"]
            del kwargs["fitline_color"]
        if "fitline_linestyle" in self.kwargs:
            self.fitline_kwargs["linestyle"] = kwargs["fitline_linestyle"]
            del kwargs["fitline_linestyle"]
        if "fitline_linewidth" in self.kwargs:
            self.fitline_kwargs["linewidth"] = kwargs["fitline_linewidth"]
            del kwargs["fitline_linewidth"]
        if "fitline_label" in self.kwargs:
            self.fitline_kwargs["label"] = kwargs["fitline_label"]
            del kwargs["fitline_label"]

        # Sets plots defaults
        self.kwargs = self.set_data_plot_default(**self.kwargs)
        self.fitline_kwargs = self.set_fit_plot_default(**self.fitline_kwargs)

        # setup the FigureCanvas
        self.fig = Figure(dpi=settings["FIG_DPI"], tight_layout=True)
        FigureCanvas.__init__(self, self.fig)
        FigureCanvas.setSizePolicy(
            self,
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Expanding,
        )
        FigureCanvas.updateGeometry(self)

        # init some statevars
        self.range_selector = None

        # create the figure and axes
        gs = self.fig.add_gridspec(3, 1)  # define three rows and one column

        # need to create ax2 first to prevent textbox related to ax1 appear behind residual plot
        self.ax2 = self.fig.add_subplot(
            gs[2, 0]
        )  # ax2 holds the plot of the residuals and spans one row
        self.ax2.title.set_text("Residual Graph")
        self.ax1 = self.fig.add_subplot(
            gs[0:2, 0], sharex=self.ax2
        )  # ax1 holds the plot of the data and spans two rows
        self.ax1.title.set_text(self.ax1_title)

        self.ax1.grid()
        self.ax2.grid()
        self.ax1.set_ylabel(
            ylabel,
            fontname=settings["TEXT_FONT"],
            fontsize=settings["TEXT_SIZE"],
        )
        self.ax2.set_ylabel(
            "residual",
            fontname=settings["TEXT_FONT"],
            fontsize=settings["TEXT_SIZE"],
        )
        self.ax2.set_xlabel(
            xlabel,
            fontname=settings["TEXT_FONT"],
            fontsize=settings["TEXT_SIZE"],
        )

        # create empty lines for the data, fit and residuals
        # (self.data_line,) = self.ax1.plot(
        #     [],
        #     [],
        #     color="black",
        #     marker="o",
        #     fillstyle="none",
        #     lw=0,
        #     label="data",
        # )
        (self.data_line,) = self.ax1.plot(
            [],
            [],
            **self.kwargs,
        )
        (self.fitted_line,) = self.ax1.plot([], [], **self.fitline_kwargs)
        (self.residual_line,) = self.ax2.plot(
            [], [], color="k", marker=".", lw=1
        )
        self.zero_res = None  # holder for a dashed hline to indicate zero in the residual plot

        # create legend
        self.ax1.legend(
            loc="best",
            fancybox=True,
            framealpha=0.5,
            prop={
                "family": settings["TEXT_FONT"],
                "size": settings["TEXT_SIZE"],
            },
        )

        # create an annotate box to hold the fitresults
        bbox_args = dict(
            boxstyle=patches.BoxStyle("round", pad=0.5), fc="0.9", alpha=0.5
        )
        self.result_box = self.ax1.annotate(
            "",
            xy=(0.5, 0.5),
            xycoords="axes fraction",
            fontname=settings["TEXT_FONT"],
            size=settings["TEXT_SIZE"],
            bbox=bbox_args,
        )
        self.result_box.draggable()

        """
        self.data.y should be complex still at this point put a print line to check,
        but set_data most likely casts to real so at this point,
        
        """
        # populate plotlines if not complex data since different plot interpretation needed for complex values (the clear circles)
        if not self.fitter.is_complex:
            self.data_line.set_data(self.data.x, self.data.y)
        else:
            self.data_line.remove()
            (self.data_line,) = self.plot_complex(
                self.data.y, "o", **self.kwargs
            )
            self.ax1.legend()

        # create errorbars if required
        if self.data.ye is not None:
            self.yerrobar = self.ax1.errorbar(
                self.data.x,
                self.data.y,
                yerr=self.data.ye,
                fmt="none",
                color=settings["BAR_Y_COLOR"],
                elinewidth=settings["BAR_Y_THICKNESS"],
                capsize=2,
            )
        if self.data.xe is not None:
            self.xerrobar = self.ax1.errorbar(
                self.data.x,
                self.data.y,
                xerr=self.data.xe,
                fmt="none",
                color=settings["BAR_X_COLOR"],
                elinewidth=settings["BAR_X_THICKNESS"],
                capsize=2,
            )

        # set the ticklabel properties
        for labels in [
            self.ax1.get_xticklabels(),
            self.ax1.get_yticklabels(),
            self.ax2.get_xticklabels(),
            self.ax2.get_yticklabels(),
        ]:
            for tick in labels:
                tick.set_color(settings["TICK_COLOR"])
                tick.set_fontproperties(settings["TICK_FONT"])
                tick.set_fontsize(settings["TICK_SIZE"])

        def add_figure_to_clipboard(event):
            if event.key == "ctrl+c":
                with io.BytesIO() as buffer:
                    self.fig.savefig(buffer)
                    QApplication.clipboard().setImage(
                        QImage.fromData(buffer.getvalue())
                    )

        self.fig.canvas.mpl_connect("key_press_event", add_figure_to_clipboard)

    def set_data_plot_default(self, **kwargs):
        kwargs["marker"] = "o" if "marker" not in kwargs else kwargs["marker"]
        kwargs["linewidth"] = (
            0 if "linewidth" not in kwargs else kwargs["linewidth"]
        )
        kwargs["label"] = "data" if "label" not in kwargs else kwargs["label"]
        return kwargs

    def set_fit_plot_default(self, **kwargs):
        kwargs["color"] = "black" if "color" not in kwargs else kwargs["color"]
        kwargs["linestyle"] = (
            "--" if "linestyle" not in kwargs else kwargs["linestyle"]
        )
        kwargs["label"] = (
            "fitted curve" if "label" not in kwargs else kwargs["label"]
        )
        return kwargs

    def set_results_box(self, text, loc):
        self.result_box.set_text(text)

        self.result_box.set_visible(True)

    def disable_results_box(self):
        self.result_box.set_visible(False)

    def toggle_rangeselector(self):
        if self.range_selector is None:
            if self.complex:
                self.range_selector = RangeSelector(
                    self.ax1, np.min(self.data.y), np.max(self.data.y)
                )
            else:
                self.range_selector = RangeSelector(
                    self.ax1, np.min(self.data.x), np.max(self.data.x)
                )
            self.redraw()
        else:
            self.range_selector.remove()
            self.range_selector = None
            self.redraw()

    def set_residuals(self, residuals):
        self.residuals = residuals

    def set_fitline(self, fitline):
        """
        Sets both the initial guess fitline and best fit line
        """
        if self.fitter.is_complex and not self.fitter.model_result:
            if not self.initial_guess_line:
                """
                Here is where changes would be made for the initial guess for complex values
                """
                (self.initial_guess_line,) = self.plot_complex(
                    fitline[1], "--g", label="initial guess"
                )
                self.ax1.legend()
        else:
            self.fitline = fitline

    def get_range(self):
        if self.range_selector is None:
            self.data.set_mask(-np.inf, np.inf)
        else:
            self.data.set_mask(*self.range_selector.get_range())

    def plot_complex(self, data, *args, **kwargs):
        """
        convenience function for plotting complex quantities
        """
        return self.ax1.plot(data.real, data.imag, *args, **kwargs)

    def create_fitline_kwargs(self, **kwargs):
        fitline_kwargs = defaultdict("str")
        if "fitline_color" in kwargs:
            fitline_kwargs["color"] = kwargs["fitline_color"]
        return fitline_kwargs

    def update_plot(self):
        # update the residuals and/or fitline if present

        if self.fitter.model_result and self.fitter.is_complex:
            fit_s21 = self.fitter.model_result.eval(
                params=self.fitter.model_result.params, x=self.fitter.data.x
            )
            # clear previous fit line
            # if self.data_line:
            #     self.data_line.remove()
            if self.fitted_line:
                self.fitted_line.remove()
            if self.initial_guess_line:
                self.initial_guess_line.remove()
                self.initial_guess_line = None

            self.ax2.lines.clear()
            # if self.complex_residuals:
            #     self.complex_residuals.cla()

            # (self.data_line,) = self.plot_complex(self.data.y, **self.kwargs)

            (self.fitted_line,) = self.plot_complex(
                fit_s21, "--k", **self.fitline_kwargs
            )
            self.complex_residuals = self.fitter.model_result.plot_residuals(
                ax=self.ax2,
                datafmt=".-k",
                parse_complex="abs",
                data_kws={"label": "residual"},
            )
            self.ax1.legend()
        else:
            if self.residuals is not None:
                # if the zero residual line is not yet created, do so
                if self.zero_res is None:
                    self.ax2.axhline(y=0, linestyle="--", color="black")

                # sort data if required
                if settings["SORT_RESIDUALS"]:
                    order = np.argsort(self.data.x)
                else:
                    order = np.arange(0, len(self.data.x))
                # print("self.data.x", self.data.x, type(self.data.x))
                # print("self.residuals", self.residuals, type(self.residuals))
                # print("order slice", order[: len(self.residuals)], type(order))
                # print("len of mask", len(self.data.mask), type(self.data.mask))
                # print("self mask", self.data.mask)

                """
                Make a copy of data.x
                iterate thru mask
                if data.mask[i] is false then delete data.x[i] in the copy
                use that copy in the residual line set data.
                """
                data_copy = np.array([])
                for i in range(len(self.data.x)):
                    if self.data.mask[i] == True:
                        data_copy = np.append(data_copy, [self.data.x[i]])
                self.residual_line.set_data(
                    data_copy[order[: len(self.residuals)]],
                    self.residuals[order[: len(self.residuals)]],
                )

                # self.residual_line.set_data(
                #     self.data.x[order],
                #     self.residuals[order[: len(self.residuals)]],
                # )

            if self.fitline is not None:
                self.fitted_line.set_data(self.fitline[0], self.fitline[1])

        # rescale the axis
        self.ax1.relim()
        self.ax1.autoscale()
        self.ax2.relim()
        self.ax2.autoscale()

        # make the min and max yscale limits of the residual plot equal
        ymax = max(np.abs(self.ax2.get_ylim()))
        self.ax2.set_ylim(-ymax, ymax)

        # draw the plot
        self.redraw()

    def redraw(self):
        # self.fig.canvas.draw()
        self.draw()


class ParamWidget(QtWidgets.QWidget):
    """Qt widget to show and change a fitparameter"""

    def __init__(self, par):
        QtWidgets.QWidget.__init__(self)
        self.par = par
        self.label = QtWidgets.QLabel(par.name)
        self.edit = QtWidgets.QLineEdit("")
        self.update_value()
        self.check = QtWidgets.QCheckBox("fix")
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self.label)
        layout.addWidget(self.edit)
        layout.addWidget(self.check)
        self.setLayout(layout)

    def read_value(self):
        """read userinput (value and fixed) in the parameter data"""
        self.par.value = float(self.edit.text())
        self.par.fixed = self.check.isChecked()
        return None

    def update_value(self):
        value = self.par.value
        self.edit.setText(float_to_str(value, settings["SIGNIFICANT_DIGITS"]))
        return None


class ReportWidget(QtWidgets.QTextEdit):
    """prints a fitreport in a non-editable textbox. Report should be a (nested) dictionary"""

    def __init__(self):
        QtWidgets.QTextEdit.__init__(
            self,
            "none",
        )
        self.setFont(
            QtGui.QFont(settings["REPORT_FONT"], settings["REPORT_SIZE"])
        )
        self.setReadOnly(True)

    def update_report(self, fitreport):
        """updates the text of the texteditbox with the content of a (nested) dictionary fitreport"""

        def print_dict(adict, level):
            for key, item in adict.items():
                if type(item) is dict:
                    if level == 1:
                        self.insertPlainText("========== ")
                    self.insertPlainText(str(key))
                    if level == 1:
                        self.insertPlainText(" ========== ")
                    self.insertPlainText("\n\n")
                    print_dict(item, level + 1)
                else:
                    if type(item) == np.float64:
                        item_str = float_to_str(
                            item, settings["SIGNIFICANT_DIGITS"]
                        )
                    else:
                        item_str = str(item)
                    self.insertPlainText(str(key) + "\t\t: " + item_str + "\n")
            self.insertPlainText("\n")

        self.clear()
        print_dict(fitreport, 1)


class ModelWidget(QtWidgets.QGroupBox):
    """Qt widget to show and control the fit model"""

    def __init__(self, model, weightoptions):
        self.model = model
        QtWidgets.QGroupBox.__init__(self, "Model settings")
        self.initGUI(weightoptions)
        self.set_weight()

    def initGUI(self, weightoptions):
        VBox = QtWidgets.QVBoxLayout()
        HBox = QtWidgets.QHBoxLayout()
        self.parviews = [ParamWidget(par) for par in self.model.fitpars]
        self.WeightLabel = QtWidgets.QLabel("Weighted Fit:")
        self.Yweightcombobox = QtWidgets.QComboBox()
        self.Yweightcombobox.addItems(weightoptions)
        HBox.addWidget(self.WeightLabel)
        HBox.addWidget(self.Yweightcombobox)
        HBox.addStretch(1)
        for parview in self.parviews:
            VBox.addWidget(parview)
        VBox.addLayout(HBox)
        self.setLayout(VBox)
        return None

    def disable_weight(self):
        self.Yweightcombobox.setDisabled(True)

    def enable_weight(self):
        self.Yweightcombobox.setEnabled(True)

    def get_weight(self):
        return self.Yweightcombobox.currentText()

    def set_weight(self):
        index = self.Yweightcombobox.findText(
            self.model.weight, QtCore.Qt.MatchFixedString
        )
        if index >= 0:
            self.Yweightcombobox.setCurrentIndex(index)

    def read_values(self):
        """reads values from userinput into the model"""
        for parview in self.parviews:
            parview.read_value()
        self.model.weight = self.get_weight()
        self.model.update()
        return None

    def update_values(self):
        for parview in self.parviews:
            parview.update_value()
        return None



# import the required packages
import random
import sys
import warnings
from inspect import signature

import numpy as np
from PyQt5 import QtCore, QtWidgets
from scipy.optimize import OptimizeWarning

from ._settings import settings
from ._tools import Fitter, value_to_string
from ._version import __version__ as CFGversion
from ._widgets import ModelWidget, PlotWidget, ReportWidget


class MainWindow(QtWidgets.QMainWindow):
    """
    Main curve fitting window with data plot, residual plot, and report
    """

    def __init__(self, afitter, xlabel, ylabel, **kwargs):
        super(MainWindow, self).__init__()

        # perform some initial default settings
        """
        We could check here if fitter.model.func is complex, then immediately call fit after initGUI() so that model.result populates 
        """
        self.fitter = afitter
        self.xlabel, self.ylabel = xlabel, ylabel
        self.output = (None, None)
        self.xerrorwarning = settings["XERRORWARNING"]
        self.initGUI(**kwargs)
        # call fit here if complex
        # if self.fitter.is_complex:
        #     self.fit()
        self.plotwidget.update_plot()

    def closeEvent(self, event):
        """needed to properly quit when running in IPython console / Spyder IDE"""
        QtWidgets.QApplication.quit()

    def initGUI(self, **kwargs):
        # main GUI proprieties
        self.setGeometry(100, 100, 1415, 900)
        self.setWindowTitle("curvefitgui " + CFGversion)
        self._main = QtWidgets.QWidget()
        self.setCentralWidget(self._main)

        """
        Holds the fit result has capability of plotting complex result, just needs to know which axis to plot on
        """
        self.model_result = self.fitter.get_model_result()

        # creating the required widgets
        self.plotwidget = PlotWidget(
            self.fitter,
            self.xlabel,
            self.ylabel,
            **kwargs,
        )  # holds the plot
        self.modelview = ModelWidget(
            self.fitter.model, self.fitter.get_weightoptions()
        )  # shows the model and allows users to set fitproperties
        self.fitbutton = QtWidgets.QPushButton("FIT", clicked=self.fit)
        self.evalbutton = QtWidgets.QPushButton(
            "INITIAL EVALUATION", clicked=self.evaluate
        )
        self.reportview = ReportWidget()  # shows the fitresults
        self.quitbutton = QtWidgets.QPushButton("QUIT", clicked=self.close)

        # create a layout for the buttons
        self.buttons = QtWidgets.QGroupBox()
        buttonslayout = QtWidgets.QHBoxLayout()
        buttonslayout.addWidget(self.evalbutton)
        buttonslayout.addWidget(self.fitbutton)
        self.buttons.setLayout(buttonslayout)

        # create a frame with a vertical layout to organize the modelview, fitbutton and reportview
        self.fitcontrolframe = QtWidgets.QGroupBox()
        fitcontrollayout = QtWidgets.QVBoxLayout()
        for widget in (
            self.modelview,
            self.buttons,
            self.reportview,
            self.quitbutton,
        ):
            fitcontrollayout.addWidget(widget)
        self.fitcontrolframe.setLayout(fitcontrollayout)

        # putting it all together: Setup the main layout
        mainlayout = QtWidgets.QHBoxLayout(self._main)
        splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        splitter.addWidget(self.plotwidget)
        splitter.addWidget(self.fitcontrolframe)
        mainlayout.addWidget(splitter)

    def showdialog(self, message, icon, info="", details=""):
        """shows an info dialog"""
        msg = QtWidgets.QMessageBox()
        if icon == "critical":
            msg.setIcon(QtWidgets.QMessageBox.Critical)
        if icon == "warning":
            msg.setIcon(QtWidgets.QMessageBox.Warning)
        msg.setText(message)
        msg.setInformativeText(info)
        msg.setWindowTitle("Message")
        msg.setDetailedText(details)
        msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
        msg.exec_()

    def set_output(self, output):
        """output should be a tuple with variables that are returned when closing the app"""
        self.output = output

    def get_output(self):
        """allows to return the currently stored output of the app when closed"""
        return self.output

    def evaluate(self):
        """updates the model and computes the model curve with the current parameter values"""
        # update the modelvalues from userinput
        try:
            self.modelview.read_values()
        except ValueError:
            self.showdialog(
                "Not a valid input initial parameter values", "critical"
            )
            return None

        # evaluate
        self.reportview.update_report({})
        self.plotwidget.canvas.set_fitline(self.fitter.get_curve())
        self.plotwidget.canvas.set_residuals(
            self.fitter.get_residuals(check=False)
        )
        self.plotwidget.canvas.disable_results_box()
        self.plotwidget.update_plot()

    def fit(self):
        """updates the model performs the fit and updates the widgets with the results"""
        # update the modelvalues from userinput
        try:
            self.modelview.read_values()
        except ValueError:
            self.showdialog(
                "Not a valid input initial parameter values", "critical"
            )
            return None

        # update fitrange
        self.plotwidget.canvas.get_range()

        # show warning on xerror data
        if (self.fitter.data.xe is not None) and self.xerrorwarning:
            self.showdialog("The error in x is ignored in the fit!", "warning")
            self.xerrorwarning = False

        # perform the fit
        with warnings.catch_warnings():
            warnings.simplefilter(
                "error", OptimizeWarning
            )  # make sure the OptimizeWarning is raised as an exception
            try:
                fitpars, fitcov, result = self.fitter.fit()
            except (ValueError, RuntimeError, OptimizeWarning):
                self.showdialog(str(sys.exc_info()[1]), "critical")

            else:
                """
                change to return result object
                """
                # update output
                self.set_output((fitpars, fitcov, result))

                # update the widgets
                self.modelview.update_values()
                self.reportview.update_report(self.fitter.get_report())
                self.plotwidget.canvas.set_fitline(self.fitter.get_fitcurve())
                self.plotwidget.canvas.set_residuals(
                    self.fitter.get_residuals()
                )
                self.plotwidget.canvas.set_results_box(
                    self._get_result_box_text(), 2
                )
                self.plotwidget.update_plot()

    def _get_result_box_text(self):
        text = "Fit results:"
        text = text + "\n" + "weight:" + self.fitter.model.weight
        for par in self.fitter.model.fitpars:
            n = par.name
            v = par.value
            e = par.sigma
            f = par.fixed
            text = text + "\n" + value_to_string(n, v, e, f)
        return text


# add is_complex parameter
def execute_gui(
    f,
    xdata,
    ydata,
    xerr,
    yerr,
    p0,
    xlabel,
    ylabel,
    absolute_sigma,
    jac,
    **kwargs
):
    """
    helper function that executes the GUI with an instance of the fitter class
    """

    if not QtWidgets.QApplication.instance():
        app = QtWidgets.QApplication([])
    else:
        app = QtWidgets.QApplication.instance()

    is_complex = False if "complex" not in kwargs else kwargs["complex"]

    showgui = True if "showgui" not in kwargs else kwargs["showgui"]

    if not showgui:
        afitter = Fitter(
            f,
            xdata,
            ydata,
            xerr,
            yerr,
            p0,
            absolute_sigma,
            jac,
            is_complex,
            **kwargs,
        )
        return afitter.fit()

    class CustomDialog(QtWidgets.QDialog):
        """
        choose fit, starting dialog that allows user to choose fit
        """

        def __init__(self):
            super().__init__()

            self.is_complex = (
                False if "complex" not in kwargs else kwargs["complex"]
            )

            def linear(x, a, b):
                """
                linear
                function: ax + b
                """

                return a * x + b

            def exp_decay(x, a, b, c):
                """
                exponential decay
                function: a * exp(-x / b) + c
                a : amplitude
                b : rate
                c : offset
                """
                return a * np.exp(-x / b) + c

            def cosine_function(x, a, b, c, d):
                """
                y = a * cos(b * x + c) + d
                """
                return a * np.cos(b * x + c) + d

            def decaying_oscillation(x, a, b, c, d, e):
                """
                y = a * exp(-x / b) * cos(c * x + d) + e
                """
                return a * np.exp(-x / b) * np.cos(c * x + d) + e

            def decaying_oscillation2(x, a, b, c, d, e):
                """
                y = a * exp(-x / b)^2 * cos(c * x + d) + e
                """
                return a * (np.exp(-x / b) ** 2) * np.cos(c * x + d) + e

            def euler(x, a):
                """
                Euler's
                y = a * exp(i * x)
                """
                return a * (np.exp(1j * (x)))

            def complex_function(x, a, b, c, d):
                """
                complex exp
                y = a * exp(i * (b * w + c)) + d
                """
                return a * np.exp(1j * (b * x + c)) + d

            self.function_map = {
                "y = ax + b": linear,
                "y = a * exp(-x / b) + c": exp_decay,
                "y = a * cos(b*x + c) + d": cosine_function,
                "y = a * exp(-x / b) * cos(c * x + d) + e": decaying_oscillation,
                "y = a * exp(-x / b)^2 * cos(c * x + d) + e": decaying_oscillation2,
                "y = a * exp(i * x)": euler,
                "y = a * exp(i * (b * w + c)) + d": complex_function,
            }
            """
            Drop down box of functions for curve fit
            """

            self.setWindowTitle("Choose Fit")
            self.button = QtWidgets.QPushButton("Start")
            self.dialog_layout = QtWidgets.QVBoxLayout()
            message = QtWidgets.QLabel("Choose Curve")

            self.combobox = QtWidgets.QComboBox()

            for item in self.function_map:
                self.combobox.addItem(item)

            # self.combobox.addItem("y = a * exp(-x / b) + c")
            # self.combobox.addItem("y = ax + b")

            self.dialog_layout.addWidget(message)
            self.dialog_layout.addWidget(self.combobox)
            self.dialog_layout.addWidget(self.button)
            self.button.clicked.connect(self.get_text)
            self.button.clicked.connect(self.close)

            self.setLayout(self.dialog_layout)

        def get_text(self):
            """
            On ok from starting pop up this function sets the fit function
            """
            self.is_complex = (
                True if "i" in self.combobox.currentText() else self.is_complex
            )
            self.func = self.function_map[self.combobox.currentText()]

    dlg = CustomDialog()
    dlg.exec_()
    if f is None:
        f = dlg.func

    """
    !!!
    From this line up to sfitter = Fitter(...) will not be needed in final product this
    is just for testing
    !!!
    """

    """
    Create Test Data when user selects function on first screen
    Start with collecting the number of fit parameters there are
    in user selected function

    """
    if ydata is None:
        sig = signature(f)
        params = sig.parameters
        num_of_params = len(params)
        random_arr = [0] * num_of_params
        for i in range(len(random_arr)):
            random_arr[i] = random.uniform(0.5, 4)

        """
        Get correct number of parameters for test data, store function values in array y
        """
        y = f(xdata, *random_arr[1:])
        print(random_arr[1:])

        """
        Create error for test data
        """
        rng = np.random.default_rng()
        test_yerr = 0.2 * np.ones_like(xdata)
        y_noise = test_yerr * rng.normal(size=xdata.size)
        ydata = y + y_noise

    """
    gonna have to add sigma to this and other places
    """
    # Here so that complex bool passes through to widgets
    if dlg.is_complex:
        kwargs["complex"] = True

    afitter = Fitter(
        f,
        xdata,
        ydata,
        xerr,
        yerr,
        p0,
        absolute_sigma,
        jac,
        dlg.is_complex,
        **kwargs,
    )

    MyApplication = MainWindow(afitter, xlabel, ylabel, **kwargs)
    MyApplication.show()
    app.exec_()
    return MyApplication.get_output()



import numpy as np

from ._gui import execute_gui


def linear_fit_gui(
    xdata,
    ydata,
    xerr=None,
    yerr=None,
    xlabel="x-axis",
    ylabel="y-axis",
    showgui=True,
):
    """
    Graphical user interface for linear fitting

    Arguments:
    ----------
    xdata : 1-D numpy array
        x-coordinates of the data
    ydata : 1-D numpy array
        y-coordinates of the data
    yerr : 1-D numpy array, optional (default:None)
        error/uncertainty in y-values used for weighted fit
        with a relative weight defined as 1/yerr**2
        (for compatibility also the use of the keyword sigma can be used for the same)
    xerr : 1-D numpy array, optional (default:None)
        error in x-values. For plotting errorbars only and ignored during fitting
    xlabel : string, optional (default:'x-values')
        x-axis title in the plot
    ylabel : string, optional (default:'y-values')
        y-axis title in the plot
    showgui : boolean, optional (default=True)
        if True, the gui is shown, otherwise not

    Returns:
    --------
    popt : numpy array
        optimal values for the fit parameters
    pcov : 2D numpy array
        the estimated covariance matrix op popt


    Notes:
    ------
    Not all functionality of curve_fit() is currently implemented. The solver is limited to the
    use of the Levenberg-Marquard algorithm.

    Examples:
    ---------

        #define x and y data as 1 dimensional numpy arrays of equal length
        xdata = np.array([1,2,3,4,5])
        ydata = np.array([-3.5, -2.4, -1, 0.5, 1.8])

        #optinally define the errors in the ydata
        yerr = np.array([0.5,0.4,0.6,0.5,0.8])

        #optionally define axis titles
        xlabel = 'time / s'
        ylabel = 'height / m'

        #execute the function
        linear_fit_gui(xdata, ydata, yerr=yerr, xlabel=xlabel, ylabel=ylabel)

    """

    # create fit function
    def f(x, a, b):
        """
        Linear fit
        y = ax + b
        a: slope
        b: intercept
        """
        return a * x + b

    p0 = None
    absolute_sigma = False
    jac = None
    kwargs = {}

    res = execute_gui(
        f,
        xdata,
        ydata,
        xerr,
        yerr,
        p0,
        xlabel,
        ylabel,
        absolute_sigma,
        jac,
        showgui,
        **kwargs,
    )
    return res


def curve_fit_gui(
    f,
    xdata,
    ydata,
    xerr=None,
    yerr=None,
    p0=None,
    xlabel="x-axis",
    ylabel="y-axis",
    absolute_sigma=False,
    jac=None,
    **kwargs,
):
    """
    Graphical user interface to the scipy curve_fit() function.

    Arguments:
    ----------
    f : callable
        function that defines the fitfunction
    xdata : 1-D numpy array
        x-coordinates of the data
    ydata : 1-D numpy array
        y-coordinates of the data
    yerr : 1-D numpy array, optional (default:None)
        error/uncertainty in y-values used for weighted fit
        with a relative weight defined as 1/yerr**2
        (for compatibility also the use of the keyword sigma can be used for the same)
    xerr : 1-D numpy array, optional (default:None)
        error in x-values. For plotting errorbars only and ignored during fitting
    xlabel : string, optional (default:'x-values')
        x-axis title in the plot
    ylabel : string, optional (default:'y-values')
        y-axis title in the plot
    p0 : array-like, optional
        initial values for fit parameters, if not specified 1 is used for each parameter
    showgui : boolean, optional (default=True)
        if True, the gui is shown, otherwise not
    absolute_sigma : boolean, optional
        see doc-string scipy.optimize.curve_fit()
    jac : callable, optional
        see doc-string scipy.optimize.curve_fit()
    kwargs
        keyword arguments for compatibility (e.g. you can use sigma to specify the error in y)


    Returns:
    --------
    popt : numpy array
        optimal values for the fit parameters
    pcov : 2D numpy array
        the estimated covariance matrix op popt

    See also:
    ---------
    scipy.optimize.curve_fit()

    Notes:
    ------
    Not all functionality of curve_fit() is currently implemented. The solver is limited to the
    use of the Levenberg-Marquard algorithm.

    Examples:
    ---------
    A minimum example is shown below

        #define a function for fitting
        def f(x, a, b):
            '''
            Linear fit
            function: y = ax + b
            a: slope
            b: intercept
            '''
            return a * x + b

        #define x and y data as 1 dimensional numpy arrays of equal length
        xdata = np.array([1, 2, 3, 4, 5])
        ydata = np.array([-3.5, -2.4, -1, 0.5, 1.8])

        #optinally define the errors in the ydata
        yerr = np.array([0.5, 0.4, 0.6, 0.5, 0.8])

        #optionally define axis titles
        xlabel = 'time / s'
        ylabel = 'height / m'

        #execute the function
        curve_fit_gui(f, xdata, ydata, yerr=yerr, xlabel=xlabel, ylabel=ylabel)

    """
    # both keyword arguments 'sigma' and 'yerr' can be used to specify errors in the ydata
    # if 'sigma' is specified, 'yerr' is ignored.
    if "sigma" in kwargs:
        yerr = kwargs["sigma"]

    res = execute_gui(
        f,
        xdata,
        ydata,
        xerr,
        yerr,
        p0,
        xlabel,
        ylabel,
        absolute_sigma,
        jac,
        **kwargs,
    )
    return res


def __main__():
    # example of use and testing"
    print("Running curve_fit_gui() with some test data taken from scipy docs")

    # define fitfunction
    def func(x, a, b, c):
        """
        exponential decay
        function: a * exp(-b * x) + c
        a : amplitude
        b : rate
        c : offset
        """
        return a * np.exp(-b * x) + c

    # create test data
    xdata = np.linspace(0, 4, 50)
    y = func(xdata, 2.5, 1.3, 0.5)
    rng = np.random.default_rng()
    yerr = 0.2 * np.ones_like(xdata)
    y_noise = yerr * rng.normal(size=xdata.size)
    ydata = y + y_noise

    # execute the gui
    popt, pcov = curve_fit_gui(
        func, xdata, ydata, yerr=yerr, xlabel="x", ylabel="y"
    )
    print(f"popt = {popt}")
    print(f"pcov = {pcov}")

    print("Running linear_fit_gui() with some test data")
    # define x and y data as 1 dimensional numpy arrays of equal length
    xdata = np.array([2, 1, 3, 4, 5])
    ydata = np.array([-2.4, -3.5, -1, 0.5, 1.8])

    # optinally define the errors in the ydata and/or xdata
    yerr = np.array([0.5, 0.4, 0.6, 0.5, 0.8])
    xerr = np.array([0.1, 0.1, 0.1, 0.1, 0.1])

    # optionally define axis titles
    xlabel = "time / s"
    ylabel = "height / m"

    # execute the function
    popt, pcov = linear_fit_gui(xdata, ydata, xerr=xerr, yerr=yerr)
    print(f"popt = {popt}")
    print(f"pcov = {pcov}")


if __name__ == "__main__":
    __main__()



import inspect
from dataclasses import dataclass, field
from typing import Any, List

import numpy as np
from lmfit import Model
from scipy import stats
from scipy.optimize import OptimizeWarning, curve_fit

from ._settings import settings


@dataclass
class FitParameter:
    """
    stores a fitparameter
    """

    name: str
    value: float = 1.0
    sigma: float = 0.0
    fixed: bool = False


@dataclass
class FitModel:
    """
    stores the model
    """

    func: Any
    jac: Any
    weight: str
    # lmfit_model: Model
    # lmfit_parameters: Any
    fitpars: List[FitParameter]
    description: str = ""

    def __post_init__(self) -> None:
        """
        Initializes lmfit model and pararemeters
        """
        self.lmfit_model = Model(self.func)
        """
        parameters and fitpars are the same, 
        however parameters is set up for lmfit, 
        whereas fitpars is connected to the PyQt frontend
        """
        self.parameters = self.lmfit_model.make_params()
        for par in self.fitpars:
            self.parameters[par.name].vary = not par.fixed
            self.parameters[par.name].value = par.value

    def evaluate(self, x):
        return self.func(x, *(par.value for par in self.fitpars))

    def get_numfitpars(self):
        return sum([not par.fixed for par in self.fitpars])

    def update(self):
        for par in self.fitpars:
            self.parameters[par.name].vary = not par.fixed
            self.parameters[par.name].value = par.value


@dataclass
class FitData:
    x: np.array  # x-data
    y: np.array  # y-data
    xe: np.array = None  # error-data on x-values
    ye: np.array = None  # error-data on y-values
    mask: List[bool] = field(init=False)

    def __post_init__(self):
        self.set_mask(-np.inf, np.inf)

    def get(self):
        result = (
            var[self.mask] if var is not None else None
            for var in [self.x, self.y, self.xe, self.ye]
        )
        return result

    def set_mask(self, xmin, xmax):
        self.mask = [xmin <= x <= xmax for x in self.x]

    def get_numfitpoints(self):
        return len(self.x[self.mask])


class Fitter:
    """class to handle the fit"""

    WEIGHTOPTIONS = ("none", "relative", "absolute")

    def __init__(
        self,
        func,
        xdata,
        ydata,
        xerr,
        yerr,
        p0,
        absolute_sigma,
        jac,
        is_complex,
        **kwargs,
    ):

        self.kwargs = kwargs
        self.data = self._init_data(xdata, ydata, xerr, yerr)
        self.model = self._init_model(func, p0, absolute_sigma, jac)
        # Model Result from lmfit
        self.model_result = None
        self.fit_is_valid = False  # becomes True a a valid fit is computed
        self.mean_squared_error = None
        self.fitreport = {}
        self.is_complex = is_complex
        self.method = "leastsq"
        if "method" in kwargs:
            self.method = kwargs["method"]

    def _init_data(self, x, y, xe, ye):

        # validate data
        for var in [x, y]:
            if type(var) is not np.ndarray:
                raise Exception("data should have type numpy array")
        if len(x) != len(y):
            raise Exception("xdata and ydata should be of equal length")

        # get error data if provided
        if ye is not None:
            if type(ye) is not np.ndarray:
                raise Exception("data should have type numpy array")
            if len(ye) != len(y):
                raise Exception("yerr and ydata should be of equal length")

        if xe is not None:
            if type(xe) is not np.ndarray:
                raise Exception("data should have type numpy array")
            if len(xe) != len(x):
                raise Exception("xerr and xdata should be of equal length")

        return FitData(x, y, xe, ye)

    def _init_model(self, func, p0, absolute_sigma, jac):
        # validate function
        if not callable(func):
            raise Exception("Not a valid fit function")

        # determine fitparameters from function
        __args = inspect.signature(func).parameters
        args = [arg.name for arg in __args.values()]

        # set initial values to 1 if not specified
        if p0 is None:
            p0 = [1.0] * len(args[1:])

        # create the fitpars
        fitpars = [
            FitParameter(arg, value) for arg, value in zip(args[1:], p0)
        ]

        # make additional modifications
        if self.data.ye is not None:
            if absolute_sigma:
                weight = self.WEIGHTOPTIONS[2]
            else:
                weight = self.WEIGHTOPTIONS[1]
        else:
            weight = self.WEIGHTOPTIONS[0]

        # create and return the FitModel class
        if func.__doc__ is None:
            description = "no info on model"
        else:
            description = strip_leading_spaces(func.__doc__)

        afitmodel = FitModel(func, jac, weight, fitpars, description)
        return afitmodel

    def fit(self):
        """
        performs the fit
        """
        # prepare model and data p0 are user guesses and pf is whether the guess is fixed or not
        p0 = [fitpar.value for fitpar in self.model.fitpars]
        pF = [fitpar.fixed for fitpar in self.model.fitpars]

        x, y, xe, ye = self.data.get()

        # check number of free fitparameters
        if self.model.get_numfitpars() == 0:
            raise OptimizeWarning(
                "There should be at least one free fitparameter"
            )

        if self._degrees_of_freedom() <= 0:
            raise OptimizeWarning(
                "The number of degrees of freedom (dof) should be at least one."
                + " Try to increase the number of datapoints or to decrease the number of free fitparameters."
            )

        absolute_sigma = self.model.weight == self.WEIGHTOPTIONS[2]
        if self.model.weight == self.WEIGHTOPTIONS[0]:
            ye = np.ones(len(y))  # error of 1 is equal to no weights

        popt, pcov, result = curve_fit_wrapper(
            self.model,
            x,
            y,
            sigma=ye,
            p0=p0,
            pF=pF,
            method=self.method,
            absolute_sigma=absolute_sigma,
            jac=self.model.jac,
        )
        # lmfit model result
        self.model_result = result

        # process results
        self.fit_is_valid = True
        stderrors = np.sqrt(np.diag(pcov))
        for fitpar, value, stderr in zip(self.model.fitpars, popt, stderrors):
            fitpar.value = value
            fitpar.sigma = stderr

        self.mean_squared_error = sum(((y - self.model.evaluate(x)) / ye) ** 2)

        self._create_report()
        return popt, pcov, result

    def get_curve(
        self, xmin=None, xmax=None, numpoints=settings["MODEL_NUMPOINTS"]
    ):
        if xmin is None:
            xmin = self.data.x.min()
        if xmax is None:
            xmax = self.data.x.max()
        xcurve = np.linspace(xmin, xmax, numpoints)
        # check to see if we have fit results yet, so program does not crash
        # in future model should be replaces with lmfit model !!!
        if not self.model_result:
            # ycurve = self.model.evaluate(xcurve)
            ycurve = self.model.lmfit_model.eval(
                self.model.parameters, x=xcurve
            )
        else:
            # lmfit
            ycurve = self.model_result.eval(self.model_result.params, x=xcurve)
        return (xcurve, ycurve)

    def get_model_result(self):
        return self.model_result

    def get_fitcurve(
        self, xmin=None, xmax=None, numpoints=settings["MODEL_NUMPOINTS"]
    ):
        if not self.fit_is_valid:
            return None
        return self.get_curve(xmin, xmax, numpoints)

    def get_residuals(self, check=True):
        """
        return the residuals as y - f(x)
        if check is True(default) only returns values if a valid fit is performed.
        """

        if not self.fit_is_valid and check:
            return None
        if not self.model_result:
            return None
        # return self.data.y - self.model.evaluate(self.data.x)
        # lmfit
        return self.model_result.residual

    def _degrees_of_freedom(self):
        return int(self.data.get_numfitpoints() - self.model.get_numfitpars())

    def _create_report(self):
        def pars_to_dict():
            parsdict = {
                par.name: dict(
                    value=par.value, stderr=par.sigma, fixed=par.fixed
                )
                for par in self.model.fitpars
            }
            return parsdict

        self.fitreport = {
            "FITPARAMETERS": {
                "model": self.model.description,
                "weight": self.model.weight,
                "N": self.data.get_numfitpoints(),
                "dof": self._degrees_of_freedom(),
                "t95-val": stats.t.ppf(0.975, self._degrees_of_freedom()),
            },
            "FITRESULTS": pars_to_dict(),
            "STATISTICS": {
                "Smin, mean squared error": self.mean_squared_error,
                "Chi Squared": self.model_result.chisqr,
                "R^2": self.model_result.rsquared,
            },
        }

    def get_report(self):
        return self.fitreport

    def get_weightoptions(self):
        if self.data.ye is not None:
            return self.WEIGHTOPTIONS
        else:
            # no error data so only one option remains
            return self.WEIGHTOPTIONS[0:1]


def curve_fit_wrapper(
    model, *pargs, p0=None, pF=None, jac=None, method="leastsq", **kwargs
):
    """
    wrapper around the scipy curve_fit() function to allow parameters to be fixed
    same call signature as the curve_fit() function except for:
    pF : 1D numpy array of size n, with n the number of fitparameters of the function
    returns the popt and cov matrices just like the original curve_fit() function
    """

    # extract arguments of the function func
    __args = inspect.signature(model.func).parameters
    args = [arg.name for arg in __args.values()]

    # populate pF (fixed marker) and p0 (param value) to default if not provided in kwargs
    if pF is None:
        pF = np.array([False for _ in args[1:]])  # set all parameters to free
    if p0 is None:
        p0 = np.array([1 for _ in args[1:]])  # set all init values to 1

    # make lists of new function arguments and function arguments to be passed to original function
    newfunc_args = [args[0]] + [
        arg for arg, fix in zip(args[1:], pF) if not fix
    ]
    orifunc_args = [args[0]] + [
        arg if not fix else str(p) for arg, fix, p in zip(args[1:], pF, p0)
    ]

    # make a string defining the new function as a lambda expression and evaluate to function
    fit_func = eval(
        f"lambda {', '.join(newfunc_args)} : func({', '.join(orifunc_args)})",
        locals(),
    )

    # make a string defining the new jacobian function (if specified) as a lambda expression and evaluate to function
    if callable(jac):
        indices = np.array(
            [index for index, value in enumerate(pF) if value == False]
        )
        fit_jac = eval(
            f"lambda {', '.join(newfunc_args)} : jac({', '.join(orifunc_args)})[:, indices]",
            locals(),
        )
    else:
        fit_jac = jac

    # populate a list of initial values for free fit-parameters
    p0_fit = np.array([p for p, fix in zip(p0, pF) if not fix])

    # peform the fit with the reduced function
    # popt, cov = curve_fit(fit_func, *pargs, p0=p0_fit, jac=fit_jac, **kwargs)

    """
    lmfit the code below uses the lmfit model, i keep the old model above as a check
    """

    # gmodel = Model(func)
    # params = gmodel.make_params()
    params = model.parameters
    for arg, fix, p in zip(args[1:], pF, p0):
        params[arg].vary = not fix
        params[arg].value = p
    # From the model wrapper class access lmfit model
    result = model.lmfit_model.fit(
        pargs[1],
        params,
        x=pargs[0],
        calc_covar=True,
        nan_policy="omit",
        method=method,
    )

    popt = np.array(list(result.best_values.values()))
    cov = result.covar

    # rebuild the popt and cov to include fixed parameters
    p0_fix = [p for p, fix in zip(p0, pF) if fix]  # values of fixed parameters
    id_fix = np.where(pF)[0]  # indices of fixed parameters
    """
    No longer needed when using lmfit model
    """
    # for id, p in zip(id_fix, p0_fix):
    #     popt = np.insert(
    #         popt, id, p, axis=0
    #     )  # fill in the popt at the free fit parameters

    # rebuild covariance matrix to include both fixed and optimized pars
    for id in id_fix:
        cov = np.insert(
            cov, id, 0, axis=1
        )  # add zero rows and columns for fixed par
        cov = np.insert(cov, id, 0, axis=0)
    # params here temporarily, in future move lmfit model to fitmodel so that params does not need to be returned here
    return popt, cov, result


def value_to_string(name, value, error, fixed):
    def get_exponent(value):
        """returns the exponent as an int generated by the :.e format specifier"""
        deci = 5
        s = f"{value:.{deci}e}"
        index_sign = s.find("e") + 1
        # Unsure why but sometimes nan prints and program crashes, this prevents it though

        if s[index_sign:] == "nan":
            return int("+00")
        return int(float(s[index_sign:]))

    def float_to_string(value, exponent, sig_digits=2):
        """
        returns a string representation in scientific notation of a number
        specified by value.
        value: the number that is converted
        exponent: the exponent used for the representation
        sig_digits: the number of significant digits that is used
        """
        deci = sig_digits + exponent - get_exponent(value) - 1
        if deci < 0:
            deci = 0
        result1 = f"{value/10**exponent:.{deci}f}"
        result2 = f"{10**exponent:.0e}"[1:]
        return result1, result2

    def to_latex(value_str, exponent, error_str=None):
        """return a latex string
        (value_str +/- error_str) x 10^(exponent)
        checks for nan or None to prevent program crash possibly
        """
        error_str = "0" if error_str is None else error_str

        if error_str and error_str != "nan":
            # != nan added to check for crash possibility
            latex_string = (
                "$= ("
                + value_str
                + "\pm"
                + error_str
                + r")\times$"
                + f"$10^{{{exponent}}}$"
            )
        else:
            latex_string = (
                "$=" + value_str + r"\times$" + f"$10^{{{exponent}}}$"
            )
        return latex_string

    x_e = get_exponent(value)
    if fixed:
        value_str, _ = float_to_string(
            value, x_e, settings["CM_SIG_DIGITS_NO_ERROR"]
        )
        error_str = None
        exponent = x_e
    else:
        dx_e = get_exponent(error)
        if x_e >= dx_e:
            value_str, _ = float_to_string(
                value, x_e, settings["CM_SIG_DIGITS"] + x_e - dx_e
            )
            error_str, _ = float_to_string(
                error, x_e, settings["CM_SIG_DIGITS"]
            )
            exponent = x_e
        else:
            value_str, _ = float_to_string(
                value, dx_e, settings["CM_SIG_DIGITS"] + x_e - dx_e
            )
            error_str, _ = float_to_string(
                error, dx_e, settings["CM_SIG_DIGITS"]
            )
            exponent = dx_e
    combined = name + to_latex(value_str, exponent, error_str)

    return combined


def float_to_str(value, digits):
    """
    return a string reps of a value in scientific notation with the number
    of significant digits specified by digits
    """
    return f"{value:1.{digits}e}"


def strip_leading_spaces(text):
    """removes leading spaces from text"""
    while text.count("\n    ") > 0:
        text = text.replace("\n    ", "\n")
    return text



import numpy as np

from curvefitgui import curve_fit_gui


class Main:
    def __init__(self):
        """
        This function are only here to create test data
        """

        def exp_decay(x, a, b, c):
            return a * np.exp(-x / b) + c

        self.xdata = np.linspace(0, 4, 50)

        # y data is only set to None because dummy data generated in backend
        # make to return fit result
        # have option to get rid of guess line
        printout = curve_fit_gui(
            None,
            self.xdata,
            None,
            title="Complex Graph",
            xlabel="y axis",
            ylabel="Imaginary",
            fitline_color="red",
            color="black",
        )


w = Main()

# y = exp_decay(self.xdata, 2.5, 1.3, 0.5)
# rng = np.random.default_rng()
# yerr = 0.2 * np.ones_like(self.xdata)
# y_noise = yerr * rng.normal(size=self.xdata.size)
# self.ydata = y + y_noise
